// --------------------------------------------------------------------
// Rank Tree
// --------------------------------------------------------------------
//  * Construct a ranktree from an array of n elements
//  * find returns the element at index (rank) k
//  * remove removes the element at rank k
//  * size returns the current size
// --------------------------------------------------------------------

class RankTree[T](A: Array[T]) {
  
  private class Node(var element: T, var size: Int,
		     var left: Node, var right: Node) {
    def this(e: T) { this(e, 1, null, null) }
    override def toString: String = {
      val lhs = if (left == null) "" else left.toString
      val rhs = if (right == null) "" else right.toString
      lhs + element.toString + "(" + size + ") " + rhs
      // lhs + element.toString + " " + rhs
    }
    def recomputeSize() { 
      val ls = if (left == null) 0 else left.size
      val rs = if (right == null) 0 else right.size
      size = ls + 1 + rs
    }
  }

  private var root: Node = buildTree(A, 0, A.length - 1)

  // Return size of Rank Tree
  def size: Int = if (root == null) 0 else root.size

  // Find item at rank k in the tree.
  // throws exception if k is out of range
  def find(k: Int): T = { 
    require(0 <= k && k < size)
    find(k, root).element
  }
  
  // Remove element at rank k from the tree.
  def remove(k: Int) { root = remove(k, root) }

  override def toString: String = 
    "[ " + (if (root == null) "" else root.toString) + "]"
  
  // --------------------------------------------------------------------
  // Internal methods
  // --------------------------------------------------------------------

  // construct the tree
  private def buildTree(A: Array[T], lo: Int, hi: Int): Node = {
    if (hi < lo)
      null
    else if (lo == hi)
      new Node(A(lo))
    else {
      val mid = (lo + hi) / 2
      val lhs = buildTree(A, lo, mid - 1)
      val rhs = buildTree(A, mid + 1, hi)
      val t = new Node(A(mid), 1, lhs, rhs)
      t.recomputeSize()
      t
    }
  }
  
  // Find node at rank k in a subtree rooted at t.
  // Returns node containing the matched item.
  private def find(k: Int, t: Node): Node = {
    if (t == null)
      throw new IllegalArgumentException
    val leftSize = if (t.left == null) 0 else t.left.size
    if (k < leftSize)
      find(k, t.left)
    else if (k == leftSize)
      t
    else
      find(k - leftSize - 1, t.right)
  }

  // Find the node at rank 0 in the subtree rooted at t.
  private def findFirst(t: Node): Node = {
    var p = t
    while (p.left != null)
      p = p.left
    p
  }

  // Remove node at rank 0 in the subtree rooted at t.
  // Returns the new root.
  private def removeFirst(t: Node): Node = {
    require(t != null)
    if (t.left == null) {
      t.right
    } else {
      t.left = removeFirst(t.left)
      t.recomputeSize()
      t
    }
  }    

  private def remove(k: Int, t: Node): Node = {
    require(t != null)
    val leftSize = if (t.left == null) 0 else t.left.size
    if (k < leftSize)
      t.left = remove(k, t.left)
    else if (k > leftSize)
      t.right = remove(k - leftSize - 1, t.right)
    else {
      // need to remove node t
      if (t.left != null && t.right != null) { // Two children
	t.element = findFirst(t.right).element
	t.right = removeFirst(t.right)
      } else
	return if (t.left != null) t.left else t.right
    }
    t.recomputeSize()
    t
  }
}
